//
//  BFMusicSearchController.swift
//  BFRecordScreenKit
//
//  Created by Harry on 2022/3/3.
//

import Foundation
import BFUIKit
import BFMediaKit
import BFCommonKit
import BFNetRequestKit
import Alamofire


class BFMusicSearchController: BFBaseViewController {
    
    var choseAction : ((PQVoiceModel?) -> Void)?
    
    var cutActionCallback: ((PQVoiceModel) -> Void)?
    
    var loadedTimeRangesObserver : NSKeyValueObservation?

    var chosedCell : BFMusicInfoSearchCell?
    
    var searchResults = [PQVoiceModel]()
    
    var currPage = 1

    // 试听音乐
    let player:AVPlayer = {
        let p = AVPlayer(playerItem: nil)
        
        return p
    }()
    
    lazy var musicTb : UITableView = {
        let tb = UITableView(frame: .zero)
        tb.delegate = self
        tb.dataSource = self
        tb.separatorColor = UIColor.hexColor(hexadecimal: "#272727")
        tb.register(BFMusicInfoSearchCell.self, forCellReuseIdentifier: "BFMusicInfoSearchCell")
        tb.tableFooterView = UIView()
        tb.backgroundColor = .clear
        return tb
    }()
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        let closeBtn = UIButton()
        closeBtn.setImage(imageInRecordScreenKit(by: "search_close"), for: .normal)
        closeBtn.frame = CGRect(x: 12, y: statusBarHeight + 2, width: 40, height: 40)
        closeBtn.addTarget(self, action: #selector(closeAction), for: .touchUpInside)
        view.addSubview(closeBtn)
        
        let searchTF = UITextField(frame: CGRect(x: 62, y: statusBarHeight + 4, width: cScreenWidth - 62 - 18, height: 36))
        let leftv = UIView(frame: CGRect(x: 0, y: 0, width: 30, height: 20))
        let iv = UIImageView(frame: CGRect(x: 6, y: 0, width: 20, height: 20))
        iv.contentMode = .scaleAspectFit
        iv.image = imageInRecordScreenKit(by: "search")
        leftv.addSubview(iv)
        searchTF.leftView = leftv
        searchTF.leftViewMode = .always
        searchTF.clearButtonMode = .whileEditing
        searchTF.placeholder = "歌名/歌手名"
        searchTF.delegate = self
        searchTF.returnKeyType = .search
        searchTF.textColor = UIColor.white
        searchTF.backgroundColor = UIColor.hexColor(hexadecimal: "#1d1d1d")
        searchTF.layer.cornerRadius = 18
        view.addSubview(searchTF)

        view.addSubview(musicTb)
        musicTb.snp.makeConstraints { make in
            make.left.right.bottom.equalToSuperview()
            make.top.equalTo(searchTF.snp.bottom).offset(4)
        }
        
        loadedTimeRangesObserver = player.observe(\AVPlayer.currentItem?.loadedTimeRanges, options: [.new, .initial]) { [weak self] (player, change) in
            
            DispatchQueue.main.async {[weak self] in
                guard let wself = self else { return }
                
                guard let ranges = change.newValue as? [CMTimeRange] else { return }
                if let cell = wself.chosedCell, let totalDur = player.currentItem?.duration, totalDur.isValid,  CMTimeCompare(ranges.first?.duration ?? .zero, totalDur) >= 0{
                    if (player.currentItem?.asset as? AVURLAsset)?.url.absoluteString ?? "b" == cell.data?.musicPath ?? "a" {
                        if cell.status == .loading{
                            cell.status = .playing
                        }
                    }
                }
//                BFLog(1, message: "开始播放音乐:\(ranges.first?.start.seconds ?? 0), dur:\( ranges.first?.duration.seconds ?? 0), tot:\(player.currentItem?.duration.seconds ?? 0)")
            }
        }
        
    }
    
    @objc func closeAction(){
        self.dismiss(animated: true, completion: nil)
    }
}

extension BFMusicSearchController : UITableViewDelegate, UITableViewDataSource {
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return searchResults.count
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "BFMusicInfoSearchCell")
        
        if let cell = cell as? BFMusicInfoSearchCell{
            cell.data = searchResults[indexPath.row]

            cell.useCallback = {[weak self, weak cell] in
                guard let wself = self else { return }
                wself.player.pause()
                cell?.status = .pause
                wself.choseAction?(cell?.data)
                wself.dismiss(animated: true, completion: nil)
            }
            cell.cutCallBack = {[weak self, weak cell] in
                guard let wself = self else { return }
                wself.player.pause()
                cell?.status = .pause
                
                if let data = cell?.data {
                    wself.dismiss(animated: true) {[weak self] in
                        guard let wself = self else { return }
                        wself.cutActionCallback?(data)
                    }
                }
            }
        }
        
        cell?.selectionStyle = .none
        
        return cell!
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return 64
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        
        if let cell = tableView.cellForRow(at: indexPath) as? BFMusicInfoSearchCell {

            cell.changeSelected(true)
            
            chosedCell = cell
            
            if cell.status == .normal {
                if let urlStr = cell.data?.musicPath, let url = URL(string: urlStr){
                    BFLog(1, message: "歌曲地址: \(url)")
                    if url.absoluteString != (player.currentItem?.asset as? AVURLAsset)?.url.absoluteString ?? "b" {
                        player.replaceCurrentItem(with: AVPlayerItem(url: url))
                        cell.status = .loading
                    }else {
                        cell.status = .playing
                    }
                    player.play()
                }
            }else if cell.status == .pause{
                player.play()
            } else{
                player.pause()
                cell.status = .pause
            }
        }
    }
    
    func tableView(_ tableView: UITableView, didDeselectRowAt indexPath: IndexPath) {
        let cell = tableView.cellForRow(at: indexPath) as? BFMusicInfoSearchCell
        cell?.status = .normal
        cell?.changeSelected(false)
        
        chosedCell = nil
        player.pause()
    }
    
    // MARK: - 搜素接口
    func requestSearchMusic(keyword: String, pageNum:Int) {
        
        BFNetRequestAdaptor.postRequestData(url: PQENVUtil.shared.materialSearchApi + searchBGMMaterialUrl, parames: ["keyWord": keyword, "pageNo": pageNum, "pageSize": 20], commonParams: commonParams(), encoding: JSONEncoding.default, isJsonEncodingNormal: true) { [weak self] response, _, error, _ in
            DispatchQueue.global().async {[weak self] in
                guard let wself = self else { return }
                
                var bgmList = [PQVoiceModel]()
                if response is NSNull || response == nil {
                    DispatchQueue.main.async {
                        cShowHUB(superView: nil, msg: "搜索失败，请重试！")
                    }
                    return
                }
                let tempArr = (response as? [String: Any])?["entityList"] as? [[String: Any]]
                if tempArr != nil {
                    for item in tempArr! {
                        let tempModel = PQVoiceModel(jsonDict: item)
                        tempModel.volume = 30
                        tempModel.voiceType = VOICETYPT.BGM.rawValue
                        bgmList.append(tempModel)
                    }
                }
                if pageNum == 1 {
                    wself.searchResults.removeAll()
                }
                wself.searchResults.append(contentsOf: bgmList)
                DispatchQueue.main.async {[weak self] in
                    guard let wself = self else { return }
                    
                    if wself.searchResults.count > 0 {
                        wself.musicTb.tableFooterView = UIView()
                    }else{
                        wself.musicTb.tableFooterView = wself.createNoDataView()
                    }
                    wself.musicTb.reloadData()
                }
            }
        }
    }
    
    func createNoDataView() -> UIView {
        let back = UIView(frame: CGRect(x: 0, y: 0, width: cScreenWidth, height: 200))
        
        let l = UILabel(frame: CGRect(x: 40, y: 60, width: cScreenWidth - 80, height: 80))
        l.text = "没有搜索到相关音乐\n换个词再试试吧"
        l.textColor = UIColor.hexColor(hexadecimal: "#787878")
        l.textAlignment = .center
        l.numberOfLines = 2
        l.font = UIFont.systemFont(ofSize: 17)
        back.addSubview(l)
        
        return back
    }
}

extension BFMusicSearchController : UITextFieldDelegate {
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        if let keyword = textField.text, keyword.replacingOccurrences(of: " ", with: "", options: .literal, range: nil).count > 0 {
            currPage = 1
            requestSearchMusic(keyword: keyword, pageNum: 1)
        }else{
            cShowHUB(superView: nil, msg: "搜索不能为空")
        }
        
        textField.resignFirstResponder()
        return true
    }
    
    
}
