//
//  BFRecordScreenBaseManager.swift
//  BFRecordScreenKit
//
//  Created by 胡志强 on 2022/1/13.
//

import Foundation
import GPUImage
import BFCommonKit
import BFMediaKit


class BFRecordScreenBaseManager : NSObject{
    
    weak var dele:BFRecordScreenController?
    var recordEndCallBack : recordEndCallBack?

    weak var assetPlayer: AVPlayer?
    weak var recordPlayer : AVPlayer?
    var progreddL : UILabel?
    var progressThumV: BFVideoThumbProgressView?
    
    var recordItem : BFRecordItemModel?{
        didSet{
            progressThumV?.recordItem = recordItem
            progressThumV?.progress = .zero
        }
    }
    
    //播放背景音乐
    lazy var playBGMPlayer : BFAudioPlayer = {
         let player = BFAudioPlayer.init()
         player.delegate = self
         return player
        
    }()
    
    var playView : GPUImageView?
    var filter = GPUImageFilter()
    
    var currentAssetProgress: CMTime = .zero{
        didSet{
            self.dele?.currentAssetProgress = currentAssetProgress
//            BFLog(1, message: "currentAssetProgress:\(currentAssetProgress.seconds)")
        }
    }

    var recording = false
    //录制的音频或变音是否正在播放
    var isPlaying = false
    // MARK: -
    func resetEnv(){
        progreddL?.text = CMTimeGetSeconds(currentAssetProgress).formatDurationToHMS()
    }
    
    func startRecord(){
        if recording{
            return
        }
        
        recording = true
//        recordItem?.voiceStickers.append(<#T##newElement: PQVoiceModel##PQVoiceModel#>)
    }

    func endRecord(){
        recording = false
    }
    
    func play(){
        
        playBGM()
    }
    
    func pause(){

        playBGMPlayer.pause()
        
        // add by ak 暂停时 isPlaying 设置为 false
        isPlaying = false
    }
    
    func changeRecordMaterail(){
        
    }
    
    func deleteRecord(at: CMTime) {
        
    }
    
    func cancelRecord(){
        
        
    }
    
    func recordFinished(isSuccess:Bool){
        progressThumV?.recordItem = recordItem
    }
    
    func dragEnd(){
        
    }
    
    func clean(){
        DispatchQueue.global().async {
            GPUImageContext.sharedFramebufferCache().purgeAllUnassignedFramebuffers()
        }
        NotificationCenter.default.removeObserver(self)

    }
    
    func checkCurrentPlayStatus(){
        if let voice = recordItem?.getUsedVoices().first(where: { mod in
            (mod.startCMTime - currentAssetProgress).seconds <= 0.05 && (mod.endCMTime - currentAssetProgress).seconds > 0.05
        }){
            let su = (recordPlayer?.currentItem?.asset as? AVURLAsset)?.url.lastPathComponent != (voice.wavFilePath as NSString?)?.lastPathComponent
            if (recordPlayer?.currentItem == nil) || su || (!su && !isPlaying){
                if su { // 我需要换源地址
                    recordPlayer?.pause()
                    recordPlayer?.replaceCurrentItem(with: AVPlayerItem(url: URL(fileURLWithPath: voice.wavFilePath)))
                }
                isPlaying = true
                let tt = max(CMTime.zero, currentAssetProgress - voice.startCMTime)
                BFLog(1, message: "tt:\(tt.seconds)")
                recordPlayer?.seek(to: tt , toleranceBefore: CMTime(seconds: 1, preferredTimescale: 10000), toleranceAfter: CMTime(seconds: 1, preferredTimescale: 10000), completionHandler: {[weak self] isfinished in
                    guard let wself = self else { return }
                    
                    wself.recordPlayer?.play()

                })
            }
            
        }else{
            
            BFLog(message: "voice is nill currentAssetProgress is\(currentAssetProgress)");
        }
    }
    
    
    /// 播放背景音乐方法
    func playBGM(){
        if(dele?.bgmModel != nil){
            
            if(dele?.bgmModel?.endCMTime.seconds == 0){
                BFLog(message: "当前素材的结束时间有问题播放不成功.");
                return
            }
            
            playBGMPlayer.configPlayer(audioPathURL: URL(fileURLWithPath: dele?.bgmModel?.wavFilePath ?? ""))
            playBGMPlayer.player?.volume = Float(dele?.bgmModel?.volume ?? 100) / 100.0
            //原背景音乐可使用时长
            let bgmSeconds = ((dele?.bgmModel?.endCMTime ?? .zero) - (dele?.bgmModel?.startCMTime ?? .zero)).seconds
            //余数
            let step = currentAssetProgress.seconds.truncatingRemainder(dividingBy: bgmSeconds)
            var seekTime = (dele?.bgmModel?.startCMTime ?? .zero).seconds + step
            if(step == 0){//当时间小于音乐时长
                seekTime += currentAssetProgress.seconds
            }
            BFLog(message: "seekTime is: \(seekTime) step is:\(step)")
           
            playBGMPlayer.seekTime(seekTime)
            playBGMPlayer.play()
        }
        
    }
    
    deinit {
        clean()
        BFLog(1, message: "\(self) release")
    }
}

extension BFRecordScreenBaseManager: BFAudioPlayerDelegate{
    
    // 播放完成
    func audioPlayerDidFinish(audioPlayer:BFAudioPlayer){
        BFLog(message: "audioPlayerDidFinish")
        //背景音乐播放完成后，素材还在播放，再次调用播放背景音乐，开始播放的起点在playBGM已处理
        if(isPlaying){
            playBGM()
            BFLog(message: "进行下一次播放背景音乐")
        }
       

    }
    
    // 播放开始
    func audioPlayerStart(audioPlayer:BFAudioPlayer){
        
    }
    // 播放进度
    func playProgress(audioPlayer:BFAudioPlayer,currentTime: TimeInterval, progress: Double){

    }
    
}
