//
//  BFCollectionViewFlowLayout.swift
//  BFRecordScreenKit
//
//  Created by 胡志强 on 2022/3/1.
//

import Foundation


protocol BFFlowLayoutDelegate: NSObjectProtocol {
    func flowLayout(_ flowLayout: BFCollectionViewFlowLayout, itemHeight indexPath: IndexPath) -> CGFloat
}

class BFCollectionViewFlowLayout: UICollectionViewFlowLayout {
    weak var dele : BFFlowLayoutDelegate?
    var totalWidth = 0.0

    var layoutAttributeArray = [UICollectionViewLayoutAttributes]()
    
    override func prepare() {
        super.prepare()
        // 计算每个 Cell 的宽度
        let itemHeight = 40.0
        // Cell 数量
        let itemCount = collectionView!.numberOfItems(inSection: 0)

        // 遍历 item 计算并缓存属性
        totalWidth = 0.0
        
        for i in layoutAttributeArray.count ..< itemCount {
            let indexPath = IndexPath(item: i, section: 0)
            
            let attr = UICollectionViewLayoutAttributes(forCellWith: indexPath)
            
            // 获取动态高度
            let itemWidth = dele?.flowLayout(self, itemHeight: indexPath) ?? 0.0

            // 赋值新的位置信息
            attr.frame = CGRect(x: totalWidth, y: 0, width: itemWidth + minimumLineSpacing, height: CGFloat(itemHeight))
            // 缓存布局属性
            layoutAttributeArray.append(attr)
            totalWidth = attr.frame.maxX
        }
    }
    
    
}

extension BFCollectionViewFlowLayout {
    
    override func layoutAttributesForElements(in rect: CGRect) -> [UICollectionViewLayoutAttributes]? {
        return layoutAttributeArray.filter {
            $0.frame.intersects(rect)
        }
    }
    
    override var collectionViewContentSize: CGSize {
        return CGSize(width: totalWidth , height: collectionView!.bounds.height)
    }
}
